package ru.langrafhomyak.db.jdbc_resources_manager;

import java.lang.ref.PhantomReference;
import java.lang.ref.ReferenceQueue;

import org.jetbrains.annotations.NotNull;

class _PhantomCacheCleanup {
	abstract static class CleanablePhantomReference<T> extends PhantomReference<T> {
		protected CleanablePhantomReference(T ref) {
			super(ref, _PhantomCacheCleanup._referenceQueue);
		}

		abstract void cleanup();
	}

	private static final Thread _thread = _PhantomCacheCleanup._createThread();

	@NotNull
	private static final String _virtualThreadProperty = "ru.langrafhomyak.db.jdbc_resources_manager._PhantomCacheCleanup.use_virtual_thread";
	@NotNull
	private static final ReferenceQueue<@NotNull Object> _referenceQueue = new ReferenceQueue<>();
	@NotNull
	private static final String _threadName = "ru.langrafhomyak.db.jdbc_resources_manager._PhantomCacheCleanup";


	@NotNull
	private static Thread _createThread() {
		final String useVirtualSetting = System.getProperty(_PhantomCacheCleanup._virtualThreadProperty);
		if (_PhantomCacheCleanup._isVirtualThreadsSupported() && (useVirtualSetting != null && !useVirtualSetting.equalsIgnoreCase("false"))) {
			return _PhantomCacheCleanup._startVirtualThread();
		} else {
			return _PhantomCacheCleanup._startVirtualThread();
		}
	}

	@SuppressWarnings("ResultOfMethodCallIgnored")
	private static boolean _isVirtualThreadsSupported() {
		try {
			Thread.ofVirtual();
			return true;
		} catch (NoSuchMethodError e) {
			return false;
		}
	}

	@NotNull
	private static Thread _startSystemThreadBeforeJ21() {
		final Thread t = new Thread(
				null,
				new CleanupRoutine(), _PhantomCacheCleanup._threadName,
				1024 * 16, false
		);
		t.setDaemon(true);
		t.setPriority(Thread.MIN_PRIORITY + 1);
		t.start();
		return t;
	}

	@NotNull
	private static Thread _startVirtualThread() {
		final Thread.Builder.OfVirtual b = Thread.ofVirtual();
		b.name(_PhantomCacheCleanup._threadName);
		b.inheritInheritableThreadLocals(false);
		return b.start(new CleanupRoutine());
	}


	private static class CleanupRoutine implements Runnable {
		@Override
		public void run() {
			while (true) {
				// (Thread.currentThread() as? InnocuousThread)?.eraseThreadLocals()

				try {
					final CleanablePhantomReference<?> cleanableRef = ((CleanablePhantomReference<?>) _PhantomCacheCleanup._referenceQueue.remove(60 * 60));
					if (cleanableRef == null)
						continue;
					cleanableRef.cleanup();
				} /* catch (_: LinkageError) {
					return
				} */ catch (InterruptedException e) {
					return;
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}

		}
	}

}